"""
Created on 15/07/2014

@author: Andrew Robinson
"""
import abc

import six


@six.add_metaclass(abc.ABCMeta)
class AuthProvider(object):
    """A base class for all Auth Providers."""

    @abc.abstractproperty
    def plugin_type(self):
        """ Short string providing labelling this plugin """

    @abc.abstractmethod
    def authenticate(self, email, username, password, options):
        """
        Check that the user credentials are correct.

        NOTE: Used within auto-registration to check it is ok to register this
        user.

        :param  email: the user's email address
        :type   email: str
        :param  username: the user's username
        :type   username: str
        :param  password: the plain text password they typed
        :type   password: str
        :param  options: options provided in auth_config_file
        :type   options: dict
        :returns:   True: accept user, False: reject user and None: reject user
            and don't try any other providers.  str, str are the email and
            username to register with if accepting. The optional dict may
            contain other attributes, e.g. roles to assign when autoregistering.
        :rtype:     (bool, str, str) or (bool, str, str, dict)
        """

    @abc.abstractmethod
    def authenticate_user(self, user, password, options):
        """
        Same as authenticate() method, except an User object is provided instead
        of a username.

        NOTE: used on normal login to check authentication and update user
        details if required.

        :param  user: the user to authenticate
        :type   user: galaxy.model.User
        :param  password: the plain text password they typed
        :type   password: str
        :param  options: options provided in auth_config_file
        :type   options: dict
        :returns:   True: accept user, False: reject user and None: reject user
            and don't try any other providers
        :rtype:     bool
        """



#!/usr/bin/env python
# Modify version...
import os
import re
import subprocess
import sys
from distutils.version import StrictVersion


DEV_RELEASE = os.environ.get("DEV_RELEASE", None) == "1"
PROJECT_DIRECTORY = os.getcwd()
PROJECT_DIRECTORY_NAME = os.path.basename(os.path.abspath(PROJECT_DIRECTORY))
PROJECT_MODULE_FILENAME = "project_galaxy_%s.py" % PROJECT_DIRECTORY_NAME
PROJECT_NAME = PROJECT_DIRECTORY_NAME.replace("_", "-")


def main(argv):
    source_dir = argv[1]
    version = argv[2]
    if not DEV_RELEASE:
        old_version = StrictVersion(version)
        old_version_tuple = old_version.version
        new_version_tuple = list(old_version_tuple)
        new_version_tuple[2] = old_version_tuple[2] + 1
        new_version = ".".join(map(str, new_version_tuple))
        new_dev_version = 0
    else:
        dev_version = re.compile(r'dev([\d]+)').search(version).group(1)
        new_dev_version = int(dev_version) + 1
        new_version = version.replace("dev%s" % dev_version, "dev%s" % new_dev_version)

    history_path = os.path.join(PROJECT_DIRECTORY, "HISTORY.rst")
    if not DEV_RELEASE:
        history = open(history_path, "r").read()

        def extend(from_str, line):
            from_str += "\n"
            return history.replace(from_str, from_str + line + "\n")

        history = extend(".. to_doc", """
---------------------
%s.dev0
---------------------

""" % new_version)
        open(history_path, "w").write(history)

    mod_path = os.path.join(PROJECT_DIRECTORY, source_dir, PROJECT_MODULE_FILENAME)
    mod = open(mod_path, "r").read()
    if not DEV_RELEASE:
        mod = re.sub(r"__version__ = '[\d\.]+'",
                    "__version__ = '%s.dev0'" % new_version,
                    mod, 1)
    else:
        mod = re.sub("dev%s" % dev_version,
                    "dev%s" % new_dev_version,
                    mod, 1)
    mod = open(mod_path, "w").write(mod)
    shell(["git", "commit", "-m", "Starting work on %s %s" % (PROJECT_NAME, new_version),
           "HISTORY.rst", mod_path])


def shell(cmds, **kwds):
    p = subprocess.Popen(cmds, **kwds)
    return p.wait()


if __name__ == "__main__":
    main(sys.argv)



from galaxy.auth.providers.alwaysreject import AlwaysReject
from galaxy.auth.providers.localdb import LocalDB
from galaxy.model import User


def test_alwaysreject():
    t = AlwaysReject()
    assert t.authenticate('testmail', 'testuser', 'secret', dict()) == (None, '', '')


def test_localdb():
    user = User(email='testmail', username='tester')
    user.set_password_cleartext('test123')
    t = LocalDB()
    reject = t.authenticate_user(user, 'wrong', {'redact_username_in_logs': False})
    accept = t.authenticate_user(user, 'test123', {'redact_username_in_logs': False})
    assert reject is False
    assert accept is True
    # Password must conform to policy (length etc)
    try:
        user.set_password_cleartext('test')
    except Exception:
        pass
    else:
        raise Exception("Password policy validation failed")



"""
Created on 15/07/2014

@author: Andrew Robinson
"""

import logging

from galaxy.exceptions import ConfigurationError
from galaxy.security.validate_user_input import transform_publicname
from galaxy.util import (
    string_as_bool,
    unicodify,
)
from ..providers import AuthProvider

try:
    import ldap
except ImportError as exc:
    ldap = None
    ldap_import_exc = exc

log = logging.getLogger(__name__)


def _get_subs(d, k, params):
    if k not in d or not d[k]:
        raise ConfigurationError("Missing '%s' parameter in LDAP options" % k)
    return str(d[k]).format(**params)


def _parse_ldap_options(options_unparsed):
    # Tag is defined in the XML but is empty
    if not options_unparsed:
        return []

    ldap_options = []

    # Valid options must start with this prefix. See help(ldap)
    prefix = "OPT_"

    for opt in options_unparsed.split(","):
        try:
            key, value = opt.split("=")
        except ValueError:
            log.warning("LDAP authenticate: Invalid syntax '%s' inside <ldap-options> element. Syntax should be option1=value1,option2=value2", opt)
            continue

        if not key.startswith(prefix):
            log.warning("LDAP authenticate: Invalid LDAP option '%s'. '%s' doesn't start with prefix '%s'", opt, key, prefix)
            continue
        try:
            key = getattr(ldap, key)
        except AttributeError:
            log.warning("LDAP authenticate: Invalid LDAP option '%s'. '%s' is not available in module ldap", opt, key)
            continue
        if value.startswith(prefix):
            try:
                value = getattr(ldap, value)
            except AttributeError:
                log.warning("LDAP authenticate: Invalid LDAP option '%s'. '%s' is not available in module ldap", opt, value)
                continue
        pair = (key, value)
        log.debug("LDAP authenticate: Valid LDAP option pair '%s' -> '%s=%s'", opt, *pair)
        ldap_options.append(pair)

    return ldap_options


class LDAP(AuthProvider):

    """
    Attempts to authenticate users against an LDAP server.

    If options include search-fields then it will attempt to search LDAP for
    those fields first.  After that it will bind to LDAP with the username
    (formatted as specified).
    """
    plugin_type = 'ldap'
    role_search_option = 'auto-register-roles'

    def __init__(self):
        super(LDAP, self).__init__()
        self.auto_create_roles_or_groups = False
        self.role_search_attribute = None

    def check_config(self, username, email, options):
        ok = True
        if options.get('continue-on-failure', 'False') == 'False':
            failure_mode = None  # reject and do not continue
        else:
            failure_mode = False  # reject but continue

        if string_as_bool(options.get('login-use-username', False)):
            if not username:
                log.debug('LDAP authenticate: username must be used to login, cannot be None')
                return ok, failure_mode
        else:
            if not email:
                log.debug('LDAP authenticate: email must be used to login, cannot be None')
                return ok, failure_mode

        auto_create_roles = string_as_bool(options.get('auto-create-roles', False))
        auto_create_groups = string_as_bool(options.get('auto-create-groups', False))
        self.auto_create_roles_or_groups = auto_create_roles or auto_create_groups
        auto_assign_roles_to_groups_only = string_as_bool(options.get('auto-assign-roles-to-groups-only', False))
        if auto_assign_roles_to_groups_only and not (auto_create_roles and auto_create_groups):
            raise ConfigurationError("If 'auto-assign-roles-to-groups-only' is True, auto-create-roles and "
                                     "auto-create-groups have to be True as well.")

        self.role_search_attribute = options.get(self.role_search_option)
        if self.auto_create_roles_or_groups and self.role_search_attribute is None:
            raise ConfigurationError("If 'auto-create-roles' or 'auto-create-groups' is True, a '%s' attribute has to"
                                     " be provided." % self.role_search_option)

        return ok, failure_mode

    def ldap_search(self, email, username, options):
        config_ok, failure_mode = self.check_config(username, email, options)
        if ldap is None:
            raise RuntimeError("Failed to load LDAP module: %s", str(ldap_import_exc))

        if not config_ok:
            return failure_mode, None

        params = {'email': email, 'username': username}

        try:
            ldap_options_raw = _get_subs(options, 'ldap-options', params)
        except ConfigurationError:
            ldap_options = ()
        else:
            ldap_options = _parse_ldap_options(ldap_options_raw)

        try:
            # setup connection
            ldap.set_option(ldap.OPT_REFERRALS, 0)

            for opt in ldap_options:
                ldap.set_option(*opt)
        except Exception:
            log.exception('LDAP authenticate: set_option exception')
            return (failure_mode, None)

        if 'search-fields' in options:
            try:
                l = ldap.initialize(_get_subs(options, 'server', params))
                l.protocol_version = 3

                if 'search-user' in options:
                    l.simple_bind_s(_get_subs(options, 'search-user', params),
                                    _get_subs(options, 'search-password', params))
                else:
                    l.simple_bind_s()

                # setup search
                attributes = {_.strip().format(**params) for _ in options['search-fields'].split(',')}
                if 'search-memberof-filter' in options:
                    attributes.add('memberOf')
                suser = l.search_ext_s(_get_subs(options, 'search-base', params),
                    ldap.SCOPE_SUBTREE,
                    _get_subs(options, 'search-filter', params), attributes,
                    timeout=60, sizelimit=1)

                # parse results
                if suser is None or len(suser) == 0:
                    log.warning('LDAP authenticate: search returned no results')
                    return (failure_mode, None)
                dn, attrs = suser[0]
                log.debug("LDAP authenticate: dn is %s", dn)
                log.debug("LDAP authenticate: search attributes are %s", attrs)
                for attr in attributes:
                    if self.role_search_attribute and attr == self.role_search_attribute[1:-1]:  # strip curly brackets
                        # keep role names as list
                        params[self.role_search_option] = [unicodify(_) for _ in attrs[attr]]
                    elif attr == 'memberOf':
                        params[attr] = [unicodify(_) for _ in attrs[attr]]
                    elif attr in attrs:
                        params[attr] = unicodify(attrs[attr][0])
                    else:
                        params[attr] = ""

                if self.auto_create_roles_or_groups and self.role_search_option not in params:
                    raise ConfigurationError("Missing or mismatching LDAP parameters for %s. Make sure the %s is "
                                             "included in the 'search-fields'." %
                                             (self.role_search_option, self.role_search_attribute))
                params['dn'] = dn
            except Exception:
                log.exception('LDAP authenticate: search exception')
                return (failure_mode, None)

        return failure_mode, params

    def authenticate(self, email, username, password, options):
        """
        See abstract method documentation.
        """
        if not options['redact_username_in_logs']:
            log.debug("LDAP authenticate: email is %s", email)
            log.debug("LDAP authenticate: username is %s", username)

        log.debug("LDAP authenticate: options are %s", options)

        failure_mode, params = self.ldap_search(email, username, options)
        if not params:
            return failure_mode, '', ''

        # allow to skip authentication to allow for pre-populating users
        if not options.get('no_password_check', False):
            params['password'] = password
            if not self._authenticate(params, options):
                return failure_mode, '', ''

        # check whether the user is a member of a specified group/domain/...
        if 'search-memberof-filter' in options:
            search_filter = _get_subs(options, 'search-memberof-filter', params)
            if not any(search_filter in ad_node_name for ad_node_name in params['memberOf']):
                return failure_mode, '', ''

        attributes = {}
        if self.auto_create_roles_or_groups:
            attributes['roles'] = params[self.role_search_option]
        return (True,
                _get_subs(options, 'auto-register-email', params),
                transform_publicname(_get_subs(options, 'auto-register-username', params)),
                attributes)

    def _authenticate(self, params, options):
        """
        Do the actual authentication by binding as the user to check their credentials
        """
        try:
            l = ldap.initialize(_get_subs(options, 'server', params))
            l.protocol_version = 3
            bind_user = _get_subs(options, 'bind-user', params)
            bind_password = _get_subs(options, 'bind-password', params)
        except Exception:
            log.exception('LDAP authenticate: initialize exception')
            return False
        try:
            l.simple_bind_s(bind_user, bind_password)
            try:
                whoami = l.whoami_s()
            except ldap.PROTOCOL_ERROR:
                # The "Who am I?" extended operation is not supported by this LDAP server
                pass
            else:
                if whoami is None:
                    raise RuntimeError('LDAP authenticate: anonymous bind')
                if not options['redact_username_in_logs']:
                    log.debug("LDAP authenticate: whoami is %s", whoami)
        except Exception as e:
            log.info('LDAP authenticate: bind exception: %s', unicodify(e))
            return False
        log.debug('LDAP authentication successful')
        return True

    def authenticate_user(self, user, password, options):
        """
        See abstract method documentation.
        """
        return self.authenticate(user.email, user.username, password, options)[0]


class ActiveDirectory(LDAP):
    """ Effectively just an alias for LDAP auth, but may contain active directory specific
    logic in the future. """
    plugin_type = 'activedirectory'


__all__ = ('LDAP', 'ActiveDirectory')



"""
Contains implementations of the authentication logic.
"""
import logging

from galaxy.auth.util import get_authenticators, parse_auth_results
from galaxy.exceptions import Conflict
from galaxy.util import string_as_bool

log = logging.getLogger(__name__)


class AuthManager(object):

    def __init__(self, app):
        self.__app = app
        self.redact_username_in_logs = app.config.redact_username_in_logs
        self.authenticators = get_authenticators(app.config.auth_config_file, app.config.auth_config_file_set)

    def check_registration_allowed(self, email, username, password):
        """Checks if the provided email/username is allowed to register."""
        message = ''
        status = 'done'
        for provider, options in self.active_authenticators(email, username, password):
            allow_reg = _get_allow_register(options)
            if allow_reg == 'challenge':
                auth_results = provider.authenticate(email, username, password, options)
                if auth_results[0] is True:
                    break
                if auth_results[0] is None:
                    message = 'Invalid email address/username or password.'
                    status = 'error'
                    break
            elif allow_reg is True:
                break
            elif allow_reg is False:
                message = 'Account registration not required for your account.  Please simply login.'
                status = 'error'
                break
        return message, status

    def check_auto_registration(self, trans, login, password, no_password_check=False):
        """
        Checks the username/email & password using auth providers in order.
        If a match is found, returns the 'auto-register' option for that provider.
        """
        if '@' in login:
            email = login
            username = None
        else:
            email = None
            username = login
        auth_return = {
            "auto_reg": False,
            "email": "",
            "username": ""
        }
        for provider, options in self.active_authenticators(email, username, password):
            if provider is None:
                log.debug("Unable to find module: %s" % options)
            else:
                options['no_password_check'] = no_password_check
                auth_results = provider.authenticate(email, username, password, options)
                if auth_results[0] is True:
                    try:
                        auth_return = parse_auth_results(trans, auth_results, options)
                    except Conflict as conflict:
                        log.exception(conflict)
                        raise
                    return auth_return
                elif auth_results[0] is None:
                    log.debug("Login: '%s', stopping due to failed non-continue", login)
                    break  # end authentication (skip rest)
        return auth_return

    def check_password(self, user, password):
        """Checks the username/email and password using auth providers."""
        for provider, options in self.active_authenticators(user.email, user.username, password):
            if provider is None:
                log.debug("Unable to find module: %s" % options)
            else:
                auth_result = provider.authenticate_user(user, password, options)
                if auth_result is True:
                    return True  # accept user
                elif auth_result is None:
                    break  # end authentication (skip rest)
        return False

    def check_change_password(self, user, current_password):
        """Checks that auth provider allows password changes and current_password
        matches.
        """
        for provider, options in self.active_authenticators(user.email, user.username, current_password):
            if provider is None:
                log.debug("Unable to find module: %s" % options)
            else:
                auth_result = provider.authenticate_user(user, current_password, options)
                if auth_result is True:
                    if string_as_bool(options.get("allow-password-change", False)):
                        return
                    else:
                        return 'Password change not supported.'
                elif auth_result is None:
                    break  # end authentication (skip rest)
        return 'Invalid current password.'

    def active_authenticators(self, email, username, password):
        """Yields AuthProvider instances for the provided configfile that match the
        filters.
        """
        try:
            for authenticator in self.authenticators:
                filter_template = authenticator.filter_template
                if filter_template:
                    filter_str = filter_template.format(email=email, username=username, password=password)
                    passed_filter = eval(filter_str, {"__builtins__": None}, {'str': str})
                    if not passed_filter:
                        continue  # skip to next
                options = authenticator.options
                options['redact_username_in_logs'] = self.redact_username_in_logs
                yield authenticator.plugin, options
        except Exception:
            log.exception("Active Authenticators Failure")
            raise


def _get_allow_register(d):
    s = d.get('allow-register', True)
    lower_s = str(s).lower()
    if lower_s == 'challenge':
        return lower_s
    else:
        return string_as_bool(s)



#!/usr/bin/env python
# Modify version...
import datetime
import os
import re
import subprocess
import sys


DEV_RELEASE = os.environ.get("DEV_RELEASE", None) == "1"
PROJECT_DIRECTORY = os.getcwd()
PROJECT_DIRECTORY_NAME = os.path.basename(os.path.abspath(PROJECT_DIRECTORY))
PROJECT_MODULE_FILENAME = "project_galaxy_%s.py" % PROJECT_DIRECTORY_NAME
PROJECT_NAME = PROJECT_DIRECTORY_NAME.replace("_", "-")


def main(argv):
    source_dir = argv[1]
    version = argv[2]
    mod_path = os.path.join(PROJECT_DIRECTORY, source_dir, PROJECT_MODULE_FILENAME)
    if not DEV_RELEASE:
        history_path = os.path.join(PROJECT_DIRECTORY, "HISTORY.rst")
        history = open(history_path, "r").read()
        today = datetime.datetime.today()
        today_str = today.strftime('%Y-%m-%d')
        history = history.replace(".dev0", " (%s)" % today_str)
        open(history_path, "w").write(history)
        mod = open(mod_path, "r").read()
        mod = re.sub(r"__version__ = '[\d\.]*\.dev\d+'",
                    "__version__ = '%s'" % version,
                    mod)
        mod = open(mod_path, "w").write(mod)
    tag = "galaxy-%s-%s" % (PROJECT_NAME, version)
    shell(["git", "commit", "-m", "Version %s of %s (tag %s)." % (version, PROJECT_NAME, tag),
           "HISTORY.rst", mod_path])
    shell(["git", "tag", tag])


def shell(cmds, **kwds):
    p = subprocess.Popen(cmds, **kwds)
    return p.wait()


if __name__ == "__main__":
    main(sys.argv)
